import typing

from registerer.exceptions import (
    ItemAlreadyRegistered,
    ItemNotRegistered,
    RegistrationError,
)
from registerer.validators import RegistryValidator

Type = typing.TypeVar("Type")


class Registerer(typing.Generic[Type]):
    """
    A generic class that can be used to create a registrey object to register class or functions.
    With type hints support.
    """

    __registry_dict: typing.Dict[str, Type] = {}
    __item_type: typing.Optional[Type]
    __validators: typing.List[RegistryValidator] = []

    def __init__(
        self,
        item_type: typing.Optional[Type] = None,
        validators: typing.Optional[typing.List[RegistryValidator]] = None,
    ):
        self.__registry_dict: typing.Dict[str, Type] = {}
        self.__item_type = item_type
        self.__validators = [] if validators is None else validators

        for validator in self.__validators:
            if not isinstance(validator, RegistryValidator):
                raise RegistrationError(
                    "the validator items should be objects of RegistryValidator or it's children."
                )

    @property
    def registry_dict(self) -> typing.Dict[str, Type]:
        """
        get a copy of the registry dict
        """
        return self.__registry_dict.copy()

    def is_registered(self, slug: str) -> bool:
        """
        is the slug registered?
        """
        return slug in self.__registry_dict

    def get(self, slug: str) -> Type:
        """
        get the registered item by slug
        """
        if not self.is_registered(slug):
            raise ItemNotRegistered(f"The item with slug='{slug}' is not registered.")
        return self.__registry_dict[slug]

    def register(self, slug: str):
        """
        register a new function or class
        """
        if self.is_registered(slug):
            raise ItemAlreadyRegistered(
                f"There is another item already registered with slug='{slug}'."
            )

        def _wrapper_function(item: Type):
            if self.__item_type is not None and not issubclass(item, self.__item_type):
                raise TypeError(
                    f"'{item.__name__}' class should be a subclass of '{self.__item_type.__name__}'."
                )

            for validator in self.__validators:
                validator.on_register(slug, item, self.__registry_dict)

            item.slug = slug
            self.__registry_dict[slug] = item
            return item

        return _wrapper_function



import typing

from registerer.exceptions import RegistrationError


class RegistryValidator:
    """
    a utility for custom validation with the Registerer.
    you can subclass this and override the on_register method, and raise an exception if you must.
    """

    def on_register(self, slug: str, item, registry_dict: typing.Dict[str, typing.Any]):
        """
        this function will be called when registering an item

        args:
            slug (str): the slug or identifier of the item that is being registered
            item (Any): the item that is being registered (the class or function)
            registry_dict (Dict[str, Any]): dictionary containing all the items that are registered so far
        """
        raise NotImplementedError


class FunctionAnnotationValidator(RegistryValidator):
    """
    use this validator to check annotations of registered functions
    """

    def __init__(self, annotations: typing.List[typing.Tuple[str, any]]) -> None:
        self.expected_annotations = annotations

    def on_register(self, slug: str, item, registry_dict: typing.Dict[str, typing.Any]):
        if self.expected_annotations and not getattr(item, "__annotations__", None):
            raise ValueError(
                f"function '{item.__name__}' does not have any type annotations."
            )

        function_annotations = list(item.__annotations__.items())
        if self.expected_annotations != function_annotations:
            raise RegistrationError(
                f"The '{item.__name__}' function with slug='{slug}' do not match the expected annotations.\n"
                f"Expected: {self.expected_annotations}\n"
                f"Got: {function_annotations}"
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['registerer']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'registerer',
    'version': '0.2.1',
    'description': 'Everything you need to implement maintainable and easy to use registry patterns in your project.',
    'long_description': '# Fast Registry\n[![](https://img.shields.io/pypi/v/python-registerer.svg)](https://pypi.python.org/pypi/python-registerer/)\n[![](https://github.com/danialkeimasi/python-registerer/workflows/tests/badge.svg)](https://github.com/danialkeimasi/python-registerer/actions)\n[![](https://img.shields.io/github/license/danialkeimasi/python-registerer.svg)](https://github.com/danialkeimasi/python-registerer/blob/master/LICENSE)\n\nEverything you need to implement maintainable and easy to use registry patterns in your project.\n# Installation\n\n```sh\npip install fast-registry\n```\n\n# Register Classes\nRegister classes with the same interface, enforce the type check and enjoy the benefits of type hints:\n```py\nfrom registerer import Registerer\n\n\nclass Animal:\n    def talk(self) -> None:\n        raise NotImplementedError\n\n\n# create a registry that requires registered items to implement the Animal interface:\nanimal_registry = Registerer(Animal)\n\n\n@animal_registry.register("dog")\nclass Dog(Animal):\n    def talk(self) -> None:\n        return "woof"\n```\n\n\n# Register Functions\nRegister functions and benefit from the function annotations validator (optional):\n```py\nfrom registerer import Registerer, FunctionAnnotationValidator\n\ndatabase_registry = Registerer(\n    validators=[\n        FunctionAnnotationValidator(annotations=[("name", str)]),\n    ]\n)\n\n@database_registry.register("sqlite")\ndef sqlite_database_connection(name: str):\n    return f"sqlite connection {name}"\n\n```\n\n# Create Custom Validators\nBy Creating a subclass of `RegistryValidator`, you can create your own validators to check registered classes/functions if you need to.\n\n# Examples\n- [Class - Simple Type Checking](https://github.com/danialkeimasi/python-registerer/blob/main/examples/class.py)\n- [Class - Custom Validator](https://github.com/danialkeimasi/python-registerer/blob/main/examples/class-with-custom-validator.py)\n- [Function - Simple](https://github.com/danialkeimasi/python-registerer/blob/main/examples/function.py)\n- [Function - With Type Annotation Validator](https://github.com/danialkeimasi/python-registerer/blob/main/examples/function-with-validator.py)\n',
    'author': 'Danial Keimasi',
    'author_email': 'danialkeimasi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/danialkeimasi/python-registerer',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
